#include <math.h>

#include ".\asciiapp.h"
#include ".\Player.h"
#include ".\BackgroundStar.h"
#include ".\Asteroid.h"
#include ".\Shot.h"
#include ".\Ore.h"
#include ".\MotherShip.h"
#include ".\Debris.h"



AsciiApp::AsciiApp() :
  m_bShutDown( false ),
  m_pPlayer( NULL ),
  m_GameState( GS_TITLE )
{

  SetCursorVisible( false );

  memset( &m_bKeyPressed, 0, sizeof( m_bKeyPressed ) );
  memset( &m_bKeyReleased, 1, sizeof( m_bKeyReleased ) );

}



AsciiApp::~AsciiApp()
{

  ClearAllObjects();

  SetCursorVisible( true );

}



void AsciiApp::Display()
{

  // professional screen clean
  for ( int i = 0; i < 80; ++i )
  {
    for ( int j = 0; j < 25; ++j )
    {
      PutChar( i, j, ' ' );
    }
  }

  switch ( m_GameState )
  {
    case GS_TITLE:
      {
        int   iColor = COL_BROWN;

        PutChar( 30, 2, 0xdb, iColor );
        PutChar( 31, 2, 0xdb, iColor );
        PutChar( 32, 2, 0xdb, iColor );
        PutChar( 30, 3, 0xdb, iColor );
        PutChar( 30, 4, 0xdb, iColor );
        PutChar( 31, 4, 0xdb, iColor );
        PutChar( 32, 4, 0xdb, iColor );
        PutChar( 32, 5, 0xdb, iColor );
        PutChar( 30, 6, 0xdb, iColor );
        PutChar( 31, 6, 0xdb, iColor );
        PutChar( 32, 6, 0xdb, iColor );

        iColor = COL_YELLOW;

        PutChar( 34, 2, 0xdb, iColor );
        PutChar( 35, 2, 0xdb, iColor );
        PutChar( 36, 2, 0xdb, iColor );
        PutChar( 35, 3, 0xdb, iColor );
        PutChar( 35, 4, 0xdb, iColor );
        PutChar( 35, 5, 0xdb, iColor );
        PutChar( 35, 6, 0xdb, iColor );

        iColor = COL_WHITE;
        PutChar( 38, 2, 0xdb, iColor );
        PutChar( 39, 2, 0xdb, iColor );
        PutChar( 40, 2, 0xdb, iColor );
        PutChar( 38, 3, 0xdb, iColor );
        PutChar( 40, 3, 0xdb, iColor );
        PutChar( 38, 4, 0xdb, iColor );
        PutChar( 40, 4, 0xdb, iColor );
        PutChar( 38, 5, 0xdb, iColor );
        PutChar( 39, 5, 0xdb, iColor );
        PutChar( 40, 5, 0xdb, iColor );
        PutChar( 38, 6, 0xdb, iColor );
        PutChar( 40, 6, 0xdb, iColor );

        iColor = COL_YELLOW;
        PutChar( 42, 2, 0xdb, iColor );
        PutChar( 43, 2, 0xdb, iColor );
        PutChar( 44, 2, 0xdb, iColor );
        PutChar( 42, 3, 0xdb, iColor );
        PutChar( 44, 3, 0xdb, iColor );
        PutChar( 42, 4, 0xdb, iColor );
        PutChar( 43, 4, 0xdb, iColor );
        PutChar( 42, 5, 0xdb, iColor );
        PutChar( 44, 5, 0xdb, iColor );
        PutChar( 42, 6, 0xdb, iColor );
        PutChar( 44, 6, 0xdb, iColor );

        iColor = COL_BROWN;
        PutChar( 46, 2, 0xdb, iColor );
        PutChar( 47, 2, 0xdb, iColor );
        PutChar( 48, 2, 0xdb, iColor );
        PutChar( 46, 3, 0xdb, iColor );
        PutChar( 46, 4, 0xdb, iColor );
        PutChar( 47, 4, 0xdb, iColor );
        PutChar( 48, 4, 0xdb, iColor );
        PutChar( 48, 5, 0xdb, iColor );
        PutChar( 46, 6, 0xdb, iColor );
        PutChar( 47, 6, 0xdb, iColor );
        PutChar( 48, 6, 0xdb, iColor );

        WriteString( 14, 10, "GDC 3h #5 Entry from Georg \"Endurion\" Rottensteiner", COL_LIGHT_GREEN );

        WriteString( 14, 12, "      Press SPACE to Play or ESCAPE to quit now" );

        WriteString( 5, 14, "In game use the cursor keys to control your craft and SPACE to shoot your" );
        WriteString( 5, 15, "laser. Shoot asteroids and collect the ore. Return to the mother ship to" );
        WriteString( 5, 16, "refuel and sell the harvested ore." );
      }
      break;
    case GS_FLYING:
      {
        tListObjects::iterator    it( m_listObjects.begin() );
        while ( it != m_listObjects.end() )
        {
          CGameObject*  pObj( *it );

          pObj->Display( *this, m_iXOffset, m_iYOffset );

          ++it;
        }
      }
      break;
    case GS_MOTHERSHIP:
      {
        WriteString( 2,  1, "Welcome aboard, Pilot SuperFly!", COL_LIGHT_GREEN );
        WriteString( 4,  3, "You can choose:", COL_LIGHT_GREEN );

        char  szTemp[200];
        wsprintf( szTemp,     "U) Upgrade your ship                            %d credits", m_pPlayer->m_iShipLevel * 400 );
        WriteString( 5,  5, szTemp, COL_LIGHT_GREEN );
        if ( !m_pPlayer->m_bRadar )
        {
          WriteString( 5,  6, "R) Buy a radar to locate the nearest Asteroid   5000 credits", COL_LIGHT_GREEN );
        }

        WriteString( 5, 11, "SPACE) Leave the SS Nautilus", COL_LIGHT_GREEN );

        WriteString( 5, 15, "What do you want to do?", COL_LIGHT_GREEN );
      }
      break;
    case GS_GAMEOVER:
      {
        WriteString( 35, 12, "GAME  OVER", COL_LIGHT_GREEN );

        if ( m_GameOverReason == GOR_OUT_OF_FUEL )
        {
          WriteString( 15, 18, "Flying without fuel is pretty dumb. Now you can drift", COL_LIGHT_GREEN );
          WriteString( 15, 19, "on and on until your oxygen is empty...", COL_LIGHT_GREEN );
        }
        else if ( m_GameOverReason == GOR_DIED )
        {
          WriteString( 15, 18, "Tough luck, sucker! You have to expect everything when", COL_LIGHT_GREEN );
          WriteString( 13, 19, "you're in outer space! Faster reflexes might have saved you.", COL_LIGHT_GREEN );
        }

        WriteString( 23, 22, "Press SPACE to return to the Title", COL_LIGHT_GREEN );
      }
      break;
  }

  if ( m_pPlayer )
  {
    char    szTemp[200];

    wsprintf( szTemp, "Shield: %3d%%", ( m_pPlayer->m_iShield * 100 ) / m_pPlayer->m_iShieldMax );
    if ( ( m_pPlayer->m_iShield * 100 ) / m_pPlayer->m_iShieldMax < 34 )
    {
      WriteString( 0, 24, szTemp, COL_LIGHT_RED, COL_LIGHT_GREEN );
    }
    else
    {
      WriteString( 0, 24, szTemp, COL_BLACK, COL_LIGHT_GREEN );
    }

    wsprintf( szTemp, "Ore: %5d", m_pPlayer->m_iOre );
    if ( m_pPlayer->m_iOre == m_pPlayer->m_iOreMax )
    {
      WriteString( 15, 24, szTemp, COL_LIGHT_RED, COL_LIGHT_GREEN );
    }
    else
    {
      WriteString( 15, 24, szTemp, COL_BLACK, COL_LIGHT_GREEN );
    }

    wsprintf( szTemp, "Fuel: %3d%%", ( m_pPlayer->m_iFuel * 100 ) / m_pPlayer->m_iFuelMax );
    if ( m_pPlayer->m_iFuel < 100 )
    {
      WriteString( 30, 24, szTemp, COL_LIGHT_RED, COL_LIGHT_GREEN );
    }
    else
    {
      WriteString( 30, 24, szTemp, COL_BLACK, COL_LIGHT_GREEN );
    }

    wsprintf( szTemp, "Sector: %03d,%03d", m_pPlayer->m_iX / 20, m_pPlayer->m_iY / 20 );
    if ( ( m_pPlayer->m_iX / 20 <= -10 )
    ||   ( m_pPlayer->m_iX / 20 >= 10 )
    ||   ( m_pPlayer->m_iY / 20 <= -10 )
    ||   ( m_pPlayer->m_iY / 20 >= 10 ) )
    {
      WriteString( 45, 24, szTemp, COL_LIGHT_RED, COL_LIGHT_GREEN );
    }
    else
    {
      WriteString( 45, 24, szTemp, COL_BLACK, COL_LIGHT_GREEN );
    }

    wsprintf( szTemp, "Credits: %d", m_pPlayer->m_iCredits );
    WriteString( 66, 24, szTemp, COL_BLACK, COL_LIGHT_GREEN );
  }

}



void AsciiApp::OnCollision( CGameObject* pObj1, CGameObject* pObj2 )
{

  if ( pObj1->m_Type == CGameObject::GO_PLAYER )
  {
    if ( pObj2->m_Type == CGameObject::GO_ASTEROID )
    {
      pObj1->m_bRemoveMe = true;
    }
    else if ( pObj2->m_Type == CGameObject::GO_DEBRIS )
    {
      pObj2->m_bRemoveMe = true;
      m_pPlayer->m_iShield--;
      if ( m_pPlayer->m_iShield <= 0 )
      {
        // tough luck sucker
        m_pPlayer->m_bRemoveMe = true;
      }
    }
    else if ( pObj2->m_Type == CGameObject::GO_ORE )
    {
      COre*   pOre = (COre*)pObj2;

      // pick up ore
      if ( m_pPlayer->m_iOre + pOre->m_iValue <= m_pPlayer->m_iOreMax )
      {
        m_pPlayer->m_iOre += pOre->m_iValue;
        pOre->m_bRemoveMe = true;
      }
    }
    else if ( pObj2->m_Type == CGameObject::GO_MOTHERSHIP )
    {
      m_GameState = GS_MOTHERSHIP;

      // automatic sell, refuel and repair
      m_pPlayer->m_iCredits += m_pPlayer->m_iOre * 30;
      m_pPlayer->m_iOre = 0;

      m_pPlayer->m_iFuel = m_pPlayer->m_iFuelMax;
      m_pPlayer->m_iShield = m_pPlayer->m_iShieldMax;
    }
  }
  else if ( pObj2->m_Type == CGameObject::GO_SHOT )
  {
    CShot*    pShot = (CShot*)pObj2;
    if ( pObj1->m_Type == CGameObject::GO_ASTEROID )
    {
      CAsteroid*    pAst = (CAsteroid*)pObj1;

      pAst->m_iHP -= pShot->m_iValue;
      if ( pAst->m_iHP <= 0 )
      {
        pAst->m_bRemoveMe = true;

        // Create Ores
         int   iOreCount = 1 + rand() % 5;

        for ( int i = 0; i < iOreCount; ++i )
        {
          COre*   pOre = new COre( pAst->m_iX + ( rand() % 2 ) * 2 - 1,
                                   pAst->m_iY + ( rand() % 2 ) * 2 - 1,
                                   0xa9,
                                   pAst->m_iBGColor,
                                   1 + pAst->m_iOreType );
      
          m_listObjects.push_back( pOre );
        }

        if ( pAst->m_iType == 1 )
        {
          // crumbly asteroid
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY, -1, -1, pAst->m_iColor, pAst->m_iBGColor ) );
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY,  0, -1, pAst->m_iColor, pAst->m_iBGColor ) );
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY,  1, -1, pAst->m_iColor, pAst->m_iBGColor ) );
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY, -1,  0, pAst->m_iColor, pAst->m_iBGColor ) );
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY,  1,  0, pAst->m_iColor, pAst->m_iBGColor ) );
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY, -1,  1, pAst->m_iColor, pAst->m_iBGColor ) );
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY,  0,  1, pAst->m_iColor, pAst->m_iBGColor ) );
          m_listObjects.push_back( new CDebris( pAst->m_iX, pAst->m_iY,  1,  1, pAst->m_iColor, pAst->m_iBGColor ) );
        }

        // create new asteroid
        CreateRandomAsteroid();
      }
      pShot->m_bRemoveMe = true;
    }
  }

}



void AsciiApp::Update( const float fElapsedTime )
{

  switch ( m_GameState )
  {
    case GS_FLYING:
      {
        // Check for collisions
        {
          tListObjects::iterator    it( m_listObjects.begin() );
          while ( it != m_listObjects.end() )
          {
            CGameObject* pObj( *it );

            if ( ( ( pObj->m_Type != CGameObject::GO_SHOT )
            &&     ( pObj->m_Type != CGameObject::GO_PLAYER ) )
            ||   ( pObj->m_bRemoveMe ) )
            {
              // not worth checking
              ++it;
              continue;
            }
            tListObjects::iterator    it2( m_listObjects.begin() );
            while ( it2 != m_listObjects.end() )
            {
              CGameObject* pObj2( *it2 );

              if ( ( pObj2->m_Type == CGameObject::GO_INVALID )
              ||   ( pObj2->m_bRemoveMe ) )
              {
                // not worth checking
                ++it2;
                continue;
              }

              if ( sqrtf( (float)( ( pObj->m_iX - pObj2->m_iX ) * ( pObj->m_iX - pObj2->m_iX ) + 
                          ( pObj->m_iY - pObj2->m_iY ) * ( pObj->m_iY - pObj2->m_iY ) ) ) < ( pObj->Size() + pObj2->Size() ) * 0.5f )
              {
                if ( pObj->m_Type < pObj2->m_Type )
                {
                  OnCollision( pObj, pObj2 );
                }
                else
                {
                  OnCollision( pObj2, pObj );
                }
              }

              ++it2;
            }
            ++it;
          }
        }

        tListObjects::iterator    it( m_listObjects.begin() );
        while ( it != m_listObjects.end() )
        {
          CGameObject* pObj( *it );

          pObj->Update( *this );

          if ( pObj->m_bRemoveMe )
          {
            if ( m_pPlayer == pObj )
            {
              m_GameState = GS_GAMEOVER;
              m_GameOverReason = GOR_DIED;
              if ( m_pPlayer->m_iFuel == 0 )
              {
                m_GameOverReason = GOR_OUT_OF_FUEL;
              }
              m_pPlayer = NULL;
            }
            delete pObj;
            it = m_listObjects.erase( it );
            continue;
          }

          ++it;
        }

        if ( m_pPlayer )
        {
          m_iXOffset = m_pPlayer->m_iX - 39;
          m_iYOffset = m_pPlayer->m_iY - 12;
        }
      }
      break;
    case GS_MOTHERSHIP:
      if ( ( m_bKeyPressed['U'] )
      &&   ( m_bKeyReleased['U'] ) )
      {
        m_bKeyReleased['U'] = false;

        if ( m_pPlayer->m_iCredits >= m_pPlayer->m_iShipLevel * 400 )
        {
          m_pPlayer->m_iCredits -= m_pPlayer->m_iShipLevel * 400;

          // upgrading
          m_pPlayer->m_iFuelMax += 500;
          m_pPlayer->m_iShieldMax++;
          m_pPlayer->m_iOreMax += 5;

          m_pPlayer->m_iFuel = m_pPlayer->m_iFuelMax;
          m_pPlayer->m_iShield = m_pPlayer->m_iShieldMax;
        }
      }
      if ( ( m_bKeyPressed['R'] )
      &&   ( m_bKeyReleased['R'] ) )
      {
        m_bKeyReleased['R'] = false;

        if ( ( m_pPlayer->m_iCredits >= 10 )
        &&   ( !m_pPlayer->m_bRadar ) )
        {
          m_pPlayer->m_iCredits -= 10;
          m_pPlayer->m_bRadar = true;
        }
      }
      if ( ( m_bKeyPressed[VK_SPACE] )
      &&   ( m_bKeyReleased[VK_SPACE] ) )
      {
        m_bKeyReleased[VK_SPACE] = false;

        m_pPlayer->m_iX = -8;
        m_pPlayer->m_iY = 0;

        m_iXOffset = m_pPlayer->m_iX - 39;
        m_iYOffset = m_pPlayer->m_iY - 12;

        m_GameState = GS_FLYING;

        // TODO - recreate asteroids
      }
      break;
    case GS_GAMEOVER:
      if ( ( m_bKeyPressed[VK_SPACE] )
      &&   ( m_bKeyReleased[VK_SPACE] ) )
      {
        m_bKeyReleased[VK_SPACE] = false;

        m_GameState = GS_TITLE;
      }
      break;
    case GS_TITLE:
      if ( ( m_bKeyPressed[VK_SPACE] )
      &&   ( m_bKeyReleased[VK_SPACE] ) )
      {
        m_bKeyReleased[VK_SPACE] = false;

        m_GameState = GS_FLYING;

        // initialise game
        ClearAllObjects();
        // background
        m_listObjects.push_back( new CBackgroundStar( 12, 4, '.', COL_GREY ) );
        m_listObjects.push_back( new CBackgroundStar( 49, 13, '*', COL_GREY ) );


        m_listObjects.push_back( new CMotherShip() );

        // player
        m_pPlayer = new CPlayer();
        m_pPlayer->m_iX = -8;
        m_listObjects.push_back( m_pPlayer );

        m_iXOffset = m_pPlayer->m_iX - 39;
        m_iYOffset = m_pPlayer->m_iY - 12;

        // Asteroids
        for ( int i = 0; i < 150; ++i )
        {
          CreateRandomAsteroid();
        }
        CAsteroid*    pAst = new CAsteroid( -80, 30,
                                            COL_BLACK,
                                            COL_BROWN );
        m_listObjects.push_back( pAst );
      }
      break;
  }

}



void AsciiApp::OnKeyEvent( const INPUT_RECORD& inputEvent )
{

  if ( inputEvent.Event.KeyEvent.bKeyDown )
  {
    m_bKeyPressed[inputEvent.Event.KeyEvent.wVirtualKeyCode] = true;

    if ( inputEvent.Event.KeyEvent.wVirtualKeyCode == VK_ESCAPE )
    {
      m_bShutDown = true;
    }
  }
  else
  {
    m_bKeyPressed[inputEvent.Event.KeyEvent.wVirtualKeyCode] = false;
    m_bKeyReleased[inputEvent.Event.KeyEvent.wVirtualKeyCode] = true;
  }

}



int AsciiApp::Run()
{

  DWORD   dwLastTicks = GetTickCount();

  while ( !m_bShutDown )
  {
    DWORD   dwNewTicks = GetTickCount();

    Display();
    ShowBufferedScreen();
    while ( dwNewTicks - dwLastTicks >= 50 )
    {
      dwLastTicks += 50;

      ProcessInput();
      Update( 0.05f );
    }
  }

  return 0;

}



void AsciiApp::CreateRandomAsteroid()
{

  int   iX = 0;
  int   iY = 0;

  while ( ( abs( iX ) < 80 )
  &&      ( abs( iY ) < 80 ) )
  {
    iX = ( rand() % 600 ) - 300;
    iY = ( rand() % 600 ) - 300;
  }

  int   iOreType = rand() % 10;

  CAsteroid*    pAst = NULL;
  if ( iOreType <= 7 )
  {
    pAst = new CAsteroid( iX, iY, COL_BLACK, COL_BROWN );
    pAst->m_iOreType = 0;
  }
  else if ( iOreType < 9 )
  {
    pAst = new CAsteroid( iX, iY, COL_BLACK, COL_GREY );
    pAst->m_iOreType = 1;
  }
  else
  {
    pAst = new CAsteroid( iX, iY, COL_BLACK, COL_YELLOW );
    pAst->m_iOreType = 2;
  }

  iOreType = rand() % 10;

  if ( iOreType >= 8 )
  {
    // crumbly asteroid
    pAst->m_iType = 1;
  }

  m_listObjects.push_back( pAst );

}



void AsciiApp::ClearAllObjects()
{

  tListObjects::iterator    it( m_listObjects.begin() );
  while ( it != m_listObjects.end() )
  {
    delete *it;

    ++it;
  }
  m_listObjects.clear();

}