#include <string>

#include "W32Console.h"



CW32Console::CW32Console() :
  m_hScreen( NULL )
{

	m_StdOut = ::GetStdHandle(STD_OUTPUT_HANDLE);
	m_StdIn = ::GetStdHandle(STD_INPUT_HANDLE);

  m_hScreen       = CreateConsoleScreenBuffer( GENERIC_READ | GENERIC_WRITE, 0, NULL, CONSOLE_TEXTMODE_BUFFER, NULL );

  if ( ( m_StdOut == INVALID_HANDLE_VALUE )
  ||   ( m_hScreen == INVALID_HANDLE_VALUE ) )
  {
    MessageBox( NULL, "Somethings seriously wrong", "Having a bad day, ey?", 0 );
  }

  SetConsoleActiveScreenBuffer( m_hScreen );

  memset( &m_ScreenBuffer, 0, sizeof( m_ScreenBuffer ) );

}


CW32Console::~CW32Console()
{

  SetConsoleActiveScreenBuffer( m_StdOut );

  CloseHandle( m_hScreen );

}



void CW32Console::PutChar( int iX, int iY, unsigned char  cChar, unsigned char ucFGColor, unsigned char ucBGColor )
{

  if ( ( iX < 0 )
  ||   ( iX >= 80 )
  ||   ( iY < 0 )
  ||   ( iY >= 25 ) )
  {
    return;
  }
  m_ScreenBuffer[iX + iY * 80].Char.AsciiChar = cChar;
  m_ScreenBuffer[iX + iY * 80].Attributes = ( ucBGColor << 4 ) | ucFGColor;

}



void CW32Console::WriteString( int iX, int iY, const std::string& strText, unsigned char ucFGColor, unsigned char ucBGColor )
{

  for ( size_t i = 0; i < strText.length(); ++i )
  {
    PutChar( iX + (int)i, iY, (unsigned char)strText[i], ucFGColor, ucBGColor );
  }

}



void CW32Console::SetCursorVisible(bool visible)
{
	CONSOLE_CURSOR_INFO info;

	::ZeroMemory( &info, sizeof( info ) );
  info.bVisible = visible;
	info.dwSize = 1;

	::SetConsoleCursorInfo( m_StdOut, &info );

}



void CW32Console::ShowBufferedScreen()
{

  COORD       coSize;

  COORD       coPos;

  SMALL_RECT  rcSource;

  coPos.X = 0;
  coPos.Y = 0;
  coSize.X = 80;
  coSize.Y = 25;

  rcSource.Left = 0;
  rcSource.Top = 0;
  rcSource.Right = 79;
  rcSource.Bottom = 24;


  WriteConsoleOutput( m_hScreen, (CHAR_INFO*)m_ScreenBuffer, coSize, coPos, &rcSource );

}



void CW32Console::OnKeyEvent( const INPUT_RECORD& inputEvent )
{
}



void CW32Console::OnMouseEvent( const INPUT_RECORD& inputEvent )
{
}



void CW32Console::ProcessInput()
{

  DWORD numEvents = 0;

  DWORD numEventsRead = 0;
    
  GetNumberOfConsoleInputEvents( m_StdIn, &numEvents );

  if ( numEvents != 0 ) 
  {
    INPUT_RECORD *eventBuffer = new INPUT_RECORD[numEvents];

    ReadConsoleInput( m_StdIn, eventBuffer, numEvents, &numEventsRead );

    for ( DWORD i = 0; i < numEventsRead; ++i ) 
    {
      if ( eventBuffer[i].EventType == KEY_EVENT ) 
      {
        OnKeyEvent( eventBuffer[i] );
      }
      else if ( eventBuffer[i].EventType == MOUSE_EVENT ) 
      {
        OnMouseEvent( eventBuffer[i] );
      }
    }
    delete[] eventBuffer;
  }

}